from trame.internal.app import get_app_instance


def change(*_args, **_kwargs):
    """
    The @change decorator allows us to register a function so that it will be
    automatically called when any of the given list of state names gets modified.

    The decorated function is passed the full state as ``**kwargs`` when possible.
    This means you should have a method profile similar to ``fn(..., **kwargs)``

    :param _args: List of names that your function should listen to
    :type _args: `list[str]`

    >>> @change('settings')
    ... def show_settings(settings, user, **kwargs):
    ...     print(settings, "for", user)

    """
    _app = get_app_instance()
    return _app.change(*_args, **_kwargs)



def compose_callbacks(*args):
    def __fn(**kwargs):
        for _fn in args:
            if _fn is None:
                continue
            try:
                _fn(**kwargs)
            except TypeError:
                _fn()

    return __fn



from .app import activate_app, create_app, deactivate_app, get_app_instance
from .dev import main
from .server import port, start, stop
from .state import (
    change, flush_state, get_state, is_dirty, is_dirty_all, State, update_state
)
from .triggers import Controller, trigger, trigger_key
from .utils import (
    AppServerThread, base_directory, ClientWindowProcess, compose_callbacks,
    get_cli_parser, get_version, is_dunder, log_js_error, print_server_info,
    validate_key_names
)



from trame.html import AbstractElement
from trame.internal.app import get_app_instance
from trame import state

# Only available 2.8.0+
from pywebvue.modules import SizeObserver as module

# Make sure used module is available
_app = get_app_instance()
_app.enable_module(module)

class SizeObserver(AbstractElement):
    """
    Component following parent size and reporting its size at mount time and when size change.

    :param name: Variable name to update with size information
    """
    def __init__(self, _name, **kwargs):
        super().__init__("size-observer", name=_name, **kwargs)
        self._attr_names += ["name"]
        # Placeholder for state
        state[_name] = None



import trame.internal as tri

NEXT_TRIGGER_ID = 0
TRIGGER_MAP = {}


def trigger_key(_fn):
    """
    Providing a function, a generated trigger name will be returned.
    The function will return the same string for the same function.

    Parameters
    ----------
    :param _fn: Function that we would like to be able to call from the client
                side.
    :type _fn: function
    :returns: Unique name that can be used for a trigger to call that function
    :rtype: str
    """
    # Return precomputed key
    global TRIGGER_MAP, NEXT_TRIGGER_ID
    if _fn in TRIGGER_MAP:
        return TRIGGER_MAP[_fn]

    # Compute unique trigger key
    NEXT_TRIGGER_ID += 1
    key = f"trigger_{NEXT_TRIGGER_ID}"
    TRIGGER_MAP[_fn] = key

    # Register function trigger
    _app = tri.get_app_instance()
    _app.trigger(key)(_fn)

    return key



from trame.internal.app import get_app_instance
from trame import state
from trame.html import AbstractElement

from pywebvue.modules import Plotly

# Make sure used module is available
_app = get_app_instance()
_app.enable_module(Plotly)

try:
    import numpy as np
except:
    pass


def safe_data(data):
    """Replace numpy array to list()"""
    result = []
    for item in data:
        for name in ["x", "y", "z"]:
            if name in item and isinstance(item[name], (np.ndarray, np.generic)):
                item[name] = item[name].tolist()

        result.append(item)

    return result


def safe_figure(fig):
    return {
        "data": safe_data(fig["data"]),
        "layout": fig["layout"],
    }


class Plotly(AbstractElement):
    """
    Create a Plotly figure element
    """

    def __init__(self, _name, **kwargs):
        super().__init__(
            "Plotly",
            data=(f"{_name}_figure.data",),
            layout=(f"{_name}_figure.layout",),
            **kwargs,
        )
        self.__figure_key = f"{_name}_figure"
        state[self.__figure_key] = {"data": [], "layout": {}}
        self._attr_names += [
            "data",
            "layout",
            ("display_mode_bar", "displayModeBar"),
            ("scroll_zoom", "scrollZoom"),
            "editable",
            ("static_plot", "staticPlot"),
            "format",
            "filename",
            "height",
            "width",
            "scale",
            ("mode_bar_buttons_to_remove", "modeBarButtonsToRemove"),
            ("mode_bar_buttons_to_add", "modeBarButtonsToAdd"),
            "locale",
            "displaylogo",
            "responsive",
            ("double_click_delay", "doubleClickDelay"),
        ]

    def update(self, plotly_fig):
        state[self.__figure_key] = safe_figure(plotly_fig.to_plotly_json())



import sys
import argparse
import json
import os

# ----------------------------------------
# Helpers
# ----------------------------------------
script_dir = os.path.dirname(__file__)
with open(os.path.join(script_dir, "./header.py")) as header:
    module_header = header.read()

tts_sensitive_elements = ["VSelect", "VDataTable"]


def get_attributes(tag):
    if len(tag.get("attributes")) == 0:
        return None

    attributes = []
    for attribute in tag.get("attributes"):
        name = attribute.get("name")
        if "(" in name:
            entry = expand_parenthetical(name, attributes)
        else:
            entry = '"' + name.replace("-", "_") + '",'
            types = attribute.get("value", {}).get("type")
            if "function" in types:
                entry += "  # JS functions unimplemented"
            attributes.append(entry)

    joined_attributes = "\n            ".join(attributes)
    return f"""
        self._attr_names += [
            {joined_attributes}
        ]"""


def get_events(tag):
    if len(tag.get("events")) == 0:
        return None

    events = []
    for event in tag.get("events"):
        entry = event.get("name")
        if "<" in entry:
            expand_dom_events(entry, events)
        else:
            if entry in ["mouseup", "mousedown", "click"]:
                entry = f"# {entry}, #Implemented in AbstractElement parent class"
            elif "-" in entry or ":" in entry:
                _entry = entry.replace(":", "_").replace("-", "_")
                entry = f'("{_entry}", "{entry}"),'
            else:
                entry = f'"{entry}",'
            events.append(entry)

    joined_events = "\n            ".join(events)
    return f"""
        self._event_names += [
            {joined_events}
        ]"""


def get_docs(tag):
    url = tag.get("doc-url", "https://vuetifyjs.com/en/introduction/why-vuetify/")
    url = url.replace("www.", "")  # www redirects to start page

    name = tag.get("name")
    attributes = tag.get("attributes", [])
    params = ""
    for attribute in attributes:
        raw_name = attribute.get("name")
        attribute_name = raw_name.replace("-", "_")
        attribute_type = attribute.get("value", {}).get("type", "string")
        description = attribute.get("description")
        if "](" in description:
            # Hide descriptions with markdown
            description = f"See description |{name}_vuetify_link|."
        params += f"""
    :param {attribute_name}: {description}
    :type {attribute_type}:"""

    events = tag.get("events")
    event_params = ""
    for event in events:
        entry = event.get("name")
        description = event.get("description")

        # Ignore calendar events, AbstractElement events
        if not "<" in entry and entry not in ["mouseup", "mousedown", "click"]:
            entry = entry.replace(":", "_").replace("-", "_")
            event_params += f"\n    :param {entry}: {description}"

    if len(event_params):
        event_params = "\n    Events\n" + event_params

    return f"""
    \"\"\"
    Vuetify's {name} component. See more info and examples |{name}_vuetify_link|.

    .. |{name}_vuetify_link| raw:: html

        <a href="{url}" target="_blank">here</a>

    {params}
    {event_params}
    \"\"\"
    """


def expand_parenthetical(attribute, attributes):
    sizes = ["sm", "md", "lg", "xl"]
    numbers = list(range(13)) if "0" in attribute else list(range(1, 13))

    prefix = ""
    if "offset-" in attribute:
        prefix = "offset_"
    if "order-" in attribute:
        prefix = "order_"

    for size in sizes:
        for number in numbers:
            attributes.append(f'"{prefix}{size}{number}",')


def expand_dom_events(event, events):
    dom_events = [
        "click",
        "dblclick",
        "mousedown",
        "mouseenter",
        "mouseleave",
        "mousemove",
        "mouseover",
        "mouseout",
        "mouseup",
        "focus",
    ]
    for dom_event in dom_events:
        events.append(f'("{dom_event}_date", "{dom_event}:date"),')
        events.append(f'("{dom_event}_month", "{dom_event}:month"),')
        events.append(f'("{dom_event}_year", "{dom_event}:year"),')


# ----------------------------------------
# Generator
# ----------------------------------------


def generate_vuetify(input_file, output_file):
    with open(input_file) as vuetify_input:
        loaded = json.loads(vuetify_input.read())
    tags = loaded.get("contributions", {}).get("html", {}).get("tags")

    generated_module = ""
    slots_names = set()

    # Extract information and generate class definitions
    for tag in tags:
        name = tag.get("name")
        tag_name = tag.get("doc-url").replace("https://www.vuetifyjs.com/api/", "")
        attributes = get_attributes(tag)
        events = get_events(tag)
        docs = get_docs(tag)
        for slot in tag.get("slots", []):
            slots_names.add(slot.get("name"))

        class_def = f"""
class {name}(AbstractElement):
    {docs}
    def __init__(self, children=None, **kwargs):
        super().__init__("{tag_name}", children, **kwargs)"""

        if name in tts_sensitive_elements:
            class_def += "\n        self.ttsSensitive()"

        if attributes is not None:
            class_def += attributes

        if events is not None:
            class_def += events

        class_def += "\n\n"

        generated_module += class_def

    # Generate slots input
    generated_slot_names = ["slot_names = ["]
    for name in slots_names:
        generated_slot_names.append(f'    "{name}",')
    generated_slot_names.append("]")
    generated_slot_names.append("Template.slot_names.update(slot_names)")
    generated_slot_names.append("")

    with open(output_file, "w") as vuetify_module:
        vuetify_module.write(module_header)
        vuetify_module.write("\n".join(generated_slot_names))
        vuetify_module.write(generated_module)


# ----------------------------------------
# Command line interface
# ----------------------------------------


def init_argparse():
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]",
        description="Generate vuetify module for trame",
    )
    parser.add_argument("-i", "--input", required=True)
    parser.add_argument("-o", "--output")
    return parser


if __name__ == "__main__":
    parser = init_argparse()
    args = parser.parse_args()
    output = args.output or "_generated.py"

    try:
        generate_vuetify(args.input, output)
    except (FileNotFoundError, IsADirectoryError) as err:
        print(f"{sys.argv[0]}: {err.strerror}", file=sys.stderr)



def main():
    """
    This function is called when using the `trame` executable.
    trame executable aim to provide additional functionalities for
    development such as dynamically reloading the Python application
    without restarting the server or the client.

    >>> trame app.py --dev

    trame executable assume you will have a `layout` variable inside
    your main script and will provide a reload button in the footer
    of your UI so you can control, when you actually want to reprocess
    your server side changes. This is especially usefull when adjusting
    UI styles.

    This functionality is in Alpha but we aim to improve it based on
    needs and feedback from the community.
    """
    from trame.internal.app import get_app_instance
    from trame.internal.utils import log_js_error

    _app = get_app_instance()
    parser = _app.cli_parser
    parser.add_argument("script", help="The script to run")
    parser.add_argument(
        "--dev", help="Allow to dynamically reload server", action="store_true"
    )
    args, _unknown = parser.parse_known_args()

    import importlib.util

    spec = importlib.util.spec_from_file_location("app", args.script)
    app = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(app)

    # Register reload trigger
    def reload():
        print("\nReloading application...")
        _app._change_callbacks.clear()
        _app._triggers.clear()
        _app.reload_app()

        # Keep sys trame ones
        _app._triggers["server_reload"] = reload
        _app._triggers["js_error"] = log_js_error

        spec = importlib.util.spec_from_file_location("app", args.script)
        app = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(app)
        app.layout.flush_content()
        print(" > done !\n")

    _app.trigger("server_reload")(reload)

    app.layout.start(debug=args.dev)



import mpld3
from trame import state

from trame.html import AbstractElement
from trame.internal.app import get_app_instance

# Only available 2.7.0+
from pywebvue.modules import Matplotlib as module

# Make sure used module is available
_app = get_app_instance()
_app.enable_module(module)


class Figure(AbstractElement):
    """
    Create a matplotlib figure viewer element

    :param name: Variable name in state
    :param figure: Matplotlib figure to show (default: None)

    >>> component1 = Figure("figure_1", figure=fig1)

    >>> component2 = Figure("figure_2")
    >>> component2.update(fig2)
    """

    def __init__(self, _name, figure=None, **kwargs):
        super().__init__("matplotlib", **kwargs)
        self._key = _name
        self._attributes["name"] = f'name="{_name}"'
        self._attributes["spec"] = f':spec="{_name}"'
        if figure is not None:
            self.update(figure)
        else:
            # Make sure we create a state entry
            state[self._key] = None

    def update(self, figure):
        state[self._key] = mpld3.fig_to_dict(figure)



from trame.html.vtk.common import (
    VtkView,
    VtkRemoteView,
    VtkLocalView,
    VtkRemoteLocalView,
    VtkAlgorithm,
    VtkCellData,
    VtkDataArray,
    VtkFieldData,
    VtkGeometryRepresentation,
    VtkGlyphRepresentation,
    VtkMesh,
    VtkPointData,
    VtkPolyData,
    VtkReader,
    VtkShareDataset,
    use_module,
)
from pywebvue.modules import ParaView

use_module(ParaView)
